package com.jiuqi.util.csv;

import java.util.HashMap;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.text.NumberFormat;
import java.io.StringReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.InputStream;
import java.io.FileNotFoundException;
import java.io.File;
import java.nio.charset.Charset;
import java.io.Reader;

public class CsvReader
{
    private Reader inputStream;
    private String fileName;
    private UserSettings userSettings;
    private Charset charset;
    private boolean useCustomRecordDelimiter;
    private DataBuffer dataBuffer;
    private ColumnBuffer columnBuffer;
    private RawRecordBuffer rawBuffer;
    private boolean[] isQualified;
    private String rawRecord;
    private HeadersHolder headersHolder;
    private boolean startedColumn;
    private boolean startedWithQualifier;
    private boolean hasMoreData;
    private char lastLetter;
    private boolean hasReadNextLine;
    private int columnsCount;
    private long currentRecord;
    private String[] values;
    private boolean initialized;
    private boolean closed;
    public static final int ESCAPE_MODE_DOUBLED = 1;
    public static final int ESCAPE_MODE_BACKSLASH = 2;
    
    public CsvReader(final String fileName, final char delimiter, final Charset charset) throws FileNotFoundException {
        this.inputStream = null;
        this.fileName = null;
        this.userSettings = new UserSettings();
        this.charset = null;
        this.useCustomRecordDelimiter = false;
        this.dataBuffer = new DataBuffer();
        this.columnBuffer = new ColumnBuffer();
        this.rawBuffer = new RawRecordBuffer();
        this.isQualified = null;
        this.rawRecord = "";
        this.headersHolder = new HeadersHolder();
        this.startedColumn = false;
        this.startedWithQualifier = false;
        this.hasMoreData = true;
        this.lastLetter = '\0';
        this.hasReadNextLine = false;
        this.columnsCount = 0;
        this.currentRecord = 0L;
        this.values = new String[10];
        this.initialized = false;
        this.closed = false;
        if (fileName == null) {
            throw new IllegalArgumentException("Parameter fileName can not be null.");
        }
        if (charset == null) {
            throw new IllegalArgumentException("Parameter charset can not be null.");
        }
        if (!new File(fileName).exists()) {
            throw new FileNotFoundException("File " + fileName + " does not exist.");
        }
        this.fileName = fileName;
        this.userSettings.Delimiter = delimiter;
        this.charset = charset;
        this.isQualified = new boolean[this.values.length];
    }
    
    public CsvReader(final String fileName, final char delimiter) throws FileNotFoundException {
        this(fileName, delimiter, Charset.forName("ISO-8859-1"));
    }
    
    public CsvReader(final String fileName) throws FileNotFoundException {
        this(fileName, ',');
    }
    
    public CsvReader(final Reader inputStream, final char delimiter) {
        this.inputStream = null;
        this.fileName = null;
        this.userSettings = new UserSettings();
        this.charset = null;
        this.useCustomRecordDelimiter = false;
        this.dataBuffer = new DataBuffer();
        this.columnBuffer = new ColumnBuffer();
        this.rawBuffer = new RawRecordBuffer();
        this.isQualified = null;
        this.rawRecord = "";
        this.headersHolder = new HeadersHolder();
        this.startedColumn = false;
        this.startedWithQualifier = false;
        this.hasMoreData = true;
        this.lastLetter = '\0';
        this.hasReadNextLine = false;
        this.columnsCount = 0;
        this.currentRecord = 0L;
        this.values = new String[10];
        this.initialized = false;
        this.closed = false;
        if (inputStream == null) {
            throw new IllegalArgumentException("Parameter inputStream can not be null.");
        }
        this.inputStream = inputStream;
        this.userSettings.Delimiter = delimiter;
        this.initialized = true;
        this.isQualified = new boolean[this.values.length];
    }
    
    public CsvReader(final Reader inputStream) {
        this(inputStream, ',');
    }
    
    public CsvReader(final InputStream inputStream, final char delimiter, final Charset charset) {
        this(new InputStreamReader(inputStream, charset), delimiter);
    }
    
    public CsvReader(final InputStream inputStream, final Charset charset) {
        this(new InputStreamReader(inputStream, charset));
    }
    
    public boolean getCaptureRawRecord() {
        return this.userSettings.CaptureRawRecord;
    }
    
    public void setCaptureRawRecord(final boolean captureRawRecord) {
        this.userSettings.CaptureRawRecord = captureRawRecord;
    }
    
    public String getRawRecord() {
        return this.rawRecord;
    }
    
    public boolean getTrimWhitespace() {
        return this.userSettings.TrimWhitespace;
    }
    
    public void setTrimWhitespace(final boolean trimWhitespace) {
        this.userSettings.TrimWhitespace = trimWhitespace;
    }
    
    public char getDelimiter() {
        return this.userSettings.Delimiter;
    }
    
    public void setDelimiter(final char delimiter) {
        this.userSettings.Delimiter = delimiter;
    }
    
    public char getRecordDelimiter() {
        return this.userSettings.RecordDelimiter;
    }
    
    public void setRecordDelimiter(final char recordDelimiter) {
        this.useCustomRecordDelimiter = true;
        this.userSettings.RecordDelimiter = recordDelimiter;
    }
    
    public char getTextQualifier() {
        return this.userSettings.TextQualifier;
    }
    
    public void setTextQualifier(final char textQualifier) {
        this.userSettings.TextQualifier = textQualifier;
    }
    
    public boolean getUseTextQualifier() {
        return this.userSettings.UseTextQualifier;
    }
    
    public void setUseTextQualifier(final boolean useTextQualifier) {
        this.userSettings.UseTextQualifier = useTextQualifier;
    }
    
    public char getComment() {
        return this.userSettings.Comment;
    }
    
    public void setComment(final char comment) {
        this.userSettings.Comment = comment;
    }
    
    public boolean getUseComments() {
        return this.userSettings.UseComments;
    }
    
    public void setUseComments(final boolean useComments) {
        this.userSettings.UseComments = useComments;
    }
    
    public int getEscapeMode() {
        return this.userSettings.EscapeMode;
    }
    
    public void setEscapeMode(final int escapeMode) throws IllegalArgumentException {
        if (escapeMode != 1 && escapeMode != 2) {
            throw new IllegalArgumentException("Parameter escapeMode must be a valid value.");
        }
        this.userSettings.EscapeMode = escapeMode;
    }
    
    public boolean getSkipEmptyRecords() {
        return this.userSettings.SkipEmptyRecords;
    }
    
    public void setSkipEmptyRecords(final boolean skipEmptyRecords) {
        this.userSettings.SkipEmptyRecords = skipEmptyRecords;
    }
    
    public boolean getSafetySwitch() {
        return this.userSettings.SafetySwitch;
    }
    
    public void setSafetySwitch(final boolean safetySwitch) {
        this.userSettings.SafetySwitch = safetySwitch;
    }
    
    public int getColumnCount() {
        return this.columnsCount;
    }
    
    public long getCurrentRecord() {
        return this.currentRecord - 1L;
    }
    
    public int getHeaderCount() {
        return this.headersHolder.Length;
    }
    
    public String[] getHeaders() throws IOException {
        this.checkClosed();
        if (this.headersHolder.Headers == null) {
            return null;
        }
        final String[] clone = new String[this.headersHolder.Length];
        System.arraycopy(this.headersHolder.Headers, 0, clone, 0, this.headersHolder.Length);
        return clone;
    }
    
    public void setHeaders(final String[] headers) {
        this.headersHolder.Headers = headers;
        this.headersHolder.IndexByName.clear();
        if (headers != null) {
            this.headersHolder.Length = headers.length;
        }
        else {
            this.headersHolder.Length = 0;
        }
        for (int i = 0; i < this.headersHolder.Length; ++i) {
            this.headersHolder.IndexByName.put(headers[i], new Integer(i));
        }
    }
    
    public String[] getValues() throws IOException {
        this.checkClosed();
        final String[] clone = new String[this.columnsCount];
        System.arraycopy(this.values, 0, clone, 0, this.columnsCount);
        return clone;
    }
    
    public String get(final int columnIndex) throws IOException {
        this.checkClosed();
        if (columnIndex > -1 && columnIndex < this.columnsCount) {
            return this.values[columnIndex];
        }
        return "";
    }
    
    public String get(final String headerName) throws IOException {
        this.checkClosed();
        return this.get(this.getIndex(headerName));
    }
    
    public static CsvReader parse(final String data) {
        if (data == null) {
            throw new IllegalArgumentException("Parameter data can not be null.");
        }
        return new CsvReader(new StringReader(data));
    }
    
    public boolean readRecord() throws IOException {
        this.checkClosed();
        this.columnsCount = 0;
        this.rawBuffer.Position = 0;
        this.dataBuffer.LineStart = this.dataBuffer.Position;
        this.hasReadNextLine = false;
        if (this.hasMoreData) {
            do {
                if (this.dataBuffer.Position == this.dataBuffer.Count) {
                    this.checkDataLength();
                }
                else {
                    this.startedWithQualifier = false;
                    char currentLetter = this.dataBuffer.Buffer[this.dataBuffer.Position];
                    if (this.userSettings.UseTextQualifier && currentLetter == this.userSettings.TextQualifier) {
                        this.lastLetter = currentLetter;
                        this.startedColumn = true;
                        this.dataBuffer.ColumnStart = this.dataBuffer.Position + 1;
                        this.startedWithQualifier = true;
                        boolean lastLetterWasQualifier = false;
                        char escapeChar = this.userSettings.TextQualifier;
                        if (this.userSettings.EscapeMode == 2) {
                            escapeChar = '\\';
                        }
                        boolean eatingTrailingJunk = false;
                        boolean lastLetterWasEscape = false;
                        boolean readingComplexEscape = false;
                        int escape = 1;
                        int escapeLength = 0;
                        char escapeValue = '\0';
                        final DataBuffer dataBuffer = this.dataBuffer;
                        ++dataBuffer.Position;
                        do {
                            if (this.dataBuffer.Position == this.dataBuffer.Count) {
                                this.checkDataLength();
                            }
                            else {
                                currentLetter = this.dataBuffer.Buffer[this.dataBuffer.Position];
                                if (eatingTrailingJunk) {
                                    this.dataBuffer.ColumnStart = this.dataBuffer.Position + 1;
                                    if (currentLetter == this.userSettings.Delimiter) {
                                        this.endColumn();
                                    }
                                    else if ((!this.useCustomRecordDelimiter && (currentLetter == '\r' || currentLetter == '\n')) || (this.useCustomRecordDelimiter && currentLetter == this.userSettings.RecordDelimiter)) {
                                        this.endColumn();
                                        this.endRecord();
                                    }
                                }
                                else if (readingComplexEscape) {
                                    ++escapeLength;
                                    switch (escape) {
                                        case 1: {
                                            escapeValue *= '\u0010';
                                            escapeValue += hexToDec(currentLetter);
                                            if (escapeLength == 4) {
                                                readingComplexEscape = false;
                                                break;
                                            }
                                            break;
                                        }
                                        case 2: {
                                            escapeValue *= '\b';
                                            escapeValue += (char)(currentLetter - '0');
                                            if (escapeLength == 3) {
                                                readingComplexEscape = false;
                                                break;
                                            }
                                            break;
                                        }
                                        case 3: {
                                            escapeValue *= '\n';
                                            escapeValue += (char)(currentLetter - '0');
                                            if (escapeLength == 3) {
                                                readingComplexEscape = false;
                                                break;
                                            }
                                            break;
                                        }
                                        case 4: {
                                            escapeValue *= '\u0010';
                                            escapeValue += hexToDec(currentLetter);
                                            if (escapeLength == 2) {
                                                readingComplexEscape = false;
                                                break;
                                            }
                                            break;
                                        }
                                    }
                                    if (!readingComplexEscape) {
                                        this.appendLetter(escapeValue);
                                    }
                                    else {
                                        this.dataBuffer.ColumnStart = this.dataBuffer.Position + 1;
                                    }
                                }
                                else if (currentLetter == this.userSettings.TextQualifier) {
                                    if (lastLetterWasEscape) {
                                        lastLetterWasEscape = false;
                                        lastLetterWasQualifier = false;
                                    }
                                    else {
                                        this.updateCurrentValue();
                                        if (this.userSettings.EscapeMode == 1) {
                                            lastLetterWasEscape = true;
                                        }
                                        lastLetterWasQualifier = true;
                                    }
                                }
                                else if (this.userSettings.EscapeMode == 2 && lastLetterWasEscape) {
                                    switch (currentLetter) {
                                        case 'n': {
                                            this.appendLetter('\n');
                                            break;
                                        }
                                        case 'r': {
                                            this.appendLetter('\r');
                                            break;
                                        }
                                        case 't': {
                                            this.appendLetter('\t');
                                            break;
                                        }
                                        case 'b': {
                                            this.appendLetter('\b');
                                            break;
                                        }
                                        case 'f': {
                                            this.appendLetter('\f');
                                            break;
                                        }
                                        case 'e': {
                                            this.appendLetter('\u001b');
                                            break;
                                        }
                                        case 'v': {
                                            this.appendLetter('\u000b');
                                            break;
                                        }
                                        case 'a': {
                                            this.appendLetter('\u0007');
                                            break;
                                        }
                                        case '0':
                                        case '1':
                                        case '2':
                                        case '3':
                                        case '4':
                                        case '5':
                                        case '6':
                                        case '7': {
                                            escape = 2;
                                            readingComplexEscape = true;
                                            escapeLength = 1;
                                            escapeValue = (char)(currentLetter - '0');
                                            this.dataBuffer.ColumnStart = this.dataBuffer.Position + 1;
                                            break;
                                        }
                                        case 'D':
                                        case 'O':
                                        case 'U':
                                        case 'X':
                                        case 'd':
                                        case 'o':
                                        case 'u':
                                        case 'x': {
                                            switch (currentLetter) {
                                                case 'U':
                                                case 'u': {
                                                    escape = 1;
                                                    break;
                                                }
                                                case 'X':
                                                case 'x': {
                                                    escape = 4;
                                                    break;
                                                }
                                                case 'O':
                                                case 'o': {
                                                    escape = 2;
                                                    break;
                                                }
                                                case 'D':
                                                case 'd': {
                                                    escape = 3;
                                                    break;
                                                }
                                            }
                                            readingComplexEscape = true;
                                            escapeLength = 0;
                                            escapeValue = '\0';
                                            this.dataBuffer.ColumnStart = this.dataBuffer.Position + 1;
                                            break;
                                        }
                                    }
                                    lastLetterWasEscape = false;
                                }
                                else if (currentLetter == escapeChar) {
                                    this.updateCurrentValue();
                                    lastLetterWasEscape = true;
                                }
                                else if (lastLetterWasQualifier) {
                                    if (currentLetter == this.userSettings.Delimiter) {
                                        this.endColumn();
                                    }
                                    else if ((!this.useCustomRecordDelimiter && (currentLetter == '\r' || currentLetter == '\n')) || (this.useCustomRecordDelimiter && currentLetter == this.userSettings.RecordDelimiter)) {
                                        this.endColumn();
                                        this.endRecord();
                                    }
                                    else {
                                        this.dataBuffer.ColumnStart = this.dataBuffer.Position + 1;
                                        eatingTrailingJunk = true;
                                    }
                                    lastLetterWasQualifier = false;
                                }
                                this.lastLetter = currentLetter;
                                if (this.startedColumn) {
                                    final DataBuffer dataBuffer2 = this.dataBuffer;
                                    ++dataBuffer2.Position;
                                    if (this.userSettings.SafetySwitch && this.dataBuffer.Position - this.dataBuffer.ColumnStart + this.columnBuffer.Position > 100000) {
                                        this.close();
                                        throw new IOException("Maximum column length of 100,000 exceeded in column " + NumberFormat.getIntegerInstance().format(this.columnsCount) + " in record " + NumberFormat.getIntegerInstance().format(this.currentRecord) + ". Set the SafetySwitch property to false" + " if you're expecting column lengths greater than 100,000 characters to" + " avoid this error.");
                                    }
                                }
                            }
                            if (this.hasMoreData) {
                                continue;
                            }
                            break;
                        } while (this.startedColumn);
                    }
                    else if (currentLetter == this.userSettings.Delimiter) {
                        this.lastLetter = currentLetter;
                        this.endColumn();
                    }
                    else if (this.useCustomRecordDelimiter && currentLetter == this.userSettings.RecordDelimiter) {
                        if (this.startedColumn || this.columnsCount > 0 || !this.userSettings.SkipEmptyRecords) {
                            this.endColumn();
                            this.endRecord();
                        }
                        else {
                            this.dataBuffer.LineStart = this.dataBuffer.Position + 1;
                        }
                        this.lastLetter = currentLetter;
                    }
                    else if (!this.useCustomRecordDelimiter && (currentLetter == '\r' || currentLetter == '\n')) {
                        if (this.startedColumn || this.columnsCount > 0 || (!this.userSettings.SkipEmptyRecords && (currentLetter == '\r' || this.lastLetter != '\r'))) {
                            this.endColumn();
                            this.endRecord();
                        }
                        else {
                            this.dataBuffer.LineStart = this.dataBuffer.Position + 1;
                        }
                        this.lastLetter = currentLetter;
                    }
                    else if (this.userSettings.UseComments && this.columnsCount == 0 && currentLetter == this.userSettings.Comment) {
                        this.lastLetter = currentLetter;
                        this.skipLine();
                    }
                    else if (this.userSettings.TrimWhitespace && (currentLetter == ' ' || currentLetter == '\t')) {
                        this.startedColumn = true;
                        this.dataBuffer.ColumnStart = this.dataBuffer.Position + 1;
                    }
                    else {
                        this.startedColumn = true;
                        this.dataBuffer.ColumnStart = this.dataBuffer.Position;
                        boolean lastLetterWasBackslash = false;
                        boolean readingComplexEscape2 = false;
                        int escape2 = 1;
                        int escapeLength2 = 0;
                        char escapeValue2 = '\0';
                        boolean firstLoop = true;
                        do {
                            if (!firstLoop && this.dataBuffer.Position == this.dataBuffer.Count) {
                                this.checkDataLength();
                            }
                            else {
                                if (!firstLoop) {
                                    currentLetter = this.dataBuffer.Buffer[this.dataBuffer.Position];
                                }
                                if (!this.userSettings.UseTextQualifier && this.userSettings.EscapeMode == 2 && currentLetter == '\\') {
                                    if (lastLetterWasBackslash) {
                                        lastLetterWasBackslash = false;
                                    }
                                    else {
                                        this.updateCurrentValue();
                                        lastLetterWasBackslash = true;
                                    }
                                }
                                else if (readingComplexEscape2) {
                                    ++escapeLength2;
                                    switch (escape2) {
                                        case 1: {
                                            escapeValue2 *= '\u0010';
                                            escapeValue2 += hexToDec(currentLetter);
                                            if (escapeLength2 == 4) {
                                                readingComplexEscape2 = false;
                                                break;
                                            }
                                            break;
                                        }
                                        case 2: {
                                            escapeValue2 *= '\b';
                                            escapeValue2 += (char)(currentLetter - '0');
                                            if (escapeLength2 == 3) {
                                                readingComplexEscape2 = false;
                                                break;
                                            }
                                            break;
                                        }
                                        case 3: {
                                            escapeValue2 *= '\n';
                                            escapeValue2 += (char)(currentLetter - '0');
                                            if (escapeLength2 == 3) {
                                                readingComplexEscape2 = false;
                                                break;
                                            }
                                            break;
                                        }
                                        case 4: {
                                            escapeValue2 *= '\u0010';
                                            escapeValue2 += hexToDec(currentLetter);
                                            if (escapeLength2 == 2) {
                                                readingComplexEscape2 = false;
                                                break;
                                            }
                                            break;
                                        }
                                    }
                                    if (!readingComplexEscape2) {
                                        this.appendLetter(escapeValue2);
                                    }
                                    else {
                                        this.dataBuffer.ColumnStart = this.dataBuffer.Position + 1;
                                    }
                                }
                                else if (this.userSettings.EscapeMode == 2 && lastLetterWasBackslash) {
                                    switch (currentLetter) {
                                        case 'n': {
                                            this.appendLetter('\n');
                                            break;
                                        }
                                        case 'r': {
                                            this.appendLetter('\r');
                                            break;
                                        }
                                        case 't': {
                                            this.appendLetter('\t');
                                            break;
                                        }
                                        case 'b': {
                                            this.appendLetter('\b');
                                            break;
                                        }
                                        case 'f': {
                                            this.appendLetter('\f');
                                            break;
                                        }
                                        case 'e': {
                                            this.appendLetter('\u001b');
                                            break;
                                        }
                                        case 'v': {
                                            this.appendLetter('\u000b');
                                            break;
                                        }
                                        case 'a': {
                                            this.appendLetter('\u0007');
                                            break;
                                        }
                                        case '0':
                                        case '1':
                                        case '2':
                                        case '3':
                                        case '4':
                                        case '5':
                                        case '6':
                                        case '7': {
                                            escape2 = 2;
                                            readingComplexEscape2 = true;
                                            escapeLength2 = 1;
                                            escapeValue2 = (char)(currentLetter - '0');
                                            this.dataBuffer.ColumnStart = this.dataBuffer.Position + 1;
                                            break;
                                        }
                                        case 'D':
                                        case 'O':
                                        case 'U':
                                        case 'X':
                                        case 'd':
                                        case 'o':
                                        case 'u':
                                        case 'x': {
                                            switch (currentLetter) {
                                                case 'U':
                                                case 'u': {
                                                    escape2 = 1;
                                                    break;
                                                }
                                                case 'X':
                                                case 'x': {
                                                    escape2 = 4;
                                                    break;
                                                }
                                                case 'O':
                                                case 'o': {
                                                    escape2 = 2;
                                                    break;
                                                }
                                                case 'D':
                                                case 'd': {
                                                    escape2 = 3;
                                                    break;
                                                }
                                            }
                                            readingComplexEscape2 = true;
                                            escapeLength2 = 0;
                                            escapeValue2 = '\0';
                                            this.dataBuffer.ColumnStart = this.dataBuffer.Position + 1;
                                            break;
                                        }
                                    }
                                    lastLetterWasBackslash = false;
                                }
                                else if (currentLetter == this.userSettings.Delimiter) {
                                    this.endColumn();
                                }
                                else if ((!this.useCustomRecordDelimiter && (currentLetter == '\r' || currentLetter == '\n')) || (this.useCustomRecordDelimiter && currentLetter == this.userSettings.RecordDelimiter)) {
                                    this.endColumn();
                                    this.endRecord();
                                }
                                this.lastLetter = currentLetter;
                                firstLoop = false;
                                if (!this.startedColumn) {
                                    continue;
                                }
                                final DataBuffer dataBuffer3 = this.dataBuffer;
                                ++dataBuffer3.Position;
                                if (this.userSettings.SafetySwitch && this.dataBuffer.Position - this.dataBuffer.ColumnStart + this.columnBuffer.Position > 100000) {
                                    this.close();
                                    throw new IOException("Maximum column length of 100,000 exceeded in column " + NumberFormat.getIntegerInstance().format(this.columnsCount) + " in record " + NumberFormat.getIntegerInstance().format(this.currentRecord) + ". Set the SafetySwitch property to false" + " if you're expecting column lengths greater than 100,000 characters to" + " avoid this error.");
                                }
                                continue;
                            }
                        } while (this.hasMoreData && this.startedColumn);
                    }
                    if (!this.hasMoreData) {
                        continue;
                    }
                    final DataBuffer dataBuffer4 = this.dataBuffer;
                    ++dataBuffer4.Position;
                }
            } while (this.hasMoreData && !this.hasReadNextLine);
            if (this.startedColumn || this.lastLetter == this.userSettings.Delimiter) {
                this.endColumn();
                this.endRecord();
            }
        }
        if (this.userSettings.CaptureRawRecord) {
            if (this.hasMoreData) {
                if (this.rawBuffer.Position == 0) {
                    this.rawRecord = new String(this.dataBuffer.Buffer, this.dataBuffer.LineStart, this.dataBuffer.Position - this.dataBuffer.LineStart - 1);
                }
                else {
                    this.rawRecord = String.valueOf(new String(this.rawBuffer.Buffer, 0, this.rawBuffer.Position)) + new String(this.dataBuffer.Buffer, this.dataBuffer.LineStart, this.dataBuffer.Position - this.dataBuffer.LineStart - 1);
                }
            }
            else {
                this.rawRecord = new String(this.rawBuffer.Buffer, 0, this.rawBuffer.Position);
            }
        }
        else {
            this.rawRecord = "";
        }
        return this.hasReadNextLine;
    }
    
    private void checkDataLength() throws IOException {
        if (!this.initialized) {
            if (this.fileName != null) {
                this.inputStream = new BufferedReader(new InputStreamReader(new FileInputStream(this.fileName), this.charset), 4096);
            }
            this.charset = null;
            this.initialized = true;
        }
        this.updateCurrentValue();
        if (this.userSettings.CaptureRawRecord && this.dataBuffer.Count > 0) {
            if (this.rawBuffer.Buffer.length - this.rawBuffer.Position < this.dataBuffer.Count - this.dataBuffer.LineStart) {
                final int newLength = this.rawBuffer.Buffer.length + Math.max(this.dataBuffer.Count - this.dataBuffer.LineStart, this.rawBuffer.Buffer.length);
                final char[] holder = new char[newLength];
                System.arraycopy(this.rawBuffer.Buffer, 0, holder, 0, this.rawBuffer.Position);
                this.rawBuffer.Buffer = holder;
            }
            System.arraycopy(this.dataBuffer.Buffer, this.dataBuffer.LineStart, this.rawBuffer.Buffer, this.rawBuffer.Position, this.dataBuffer.Count - this.dataBuffer.LineStart);
            final RawRecordBuffer rawBuffer = this.rawBuffer;
            rawBuffer.Position += this.dataBuffer.Count - this.dataBuffer.LineStart;
        }
        try {
            this.dataBuffer.Count = this.inputStream.read(this.dataBuffer.Buffer, 0, this.dataBuffer.Buffer.length);
        }
        catch (final IOException ex) {
            this.close();
            throw ex;
        }
        if (this.dataBuffer.Count == -1) {
            this.hasMoreData = false;
        }
        this.dataBuffer.Position = 0;
        this.dataBuffer.LineStart = 0;
        this.dataBuffer.ColumnStart = 0;
    }
    
    public boolean readHeaders() throws IOException {
        final boolean result = this.readRecord();
        this.headersHolder.Length = this.columnsCount;
        this.headersHolder.Headers = new String[this.columnsCount];
        for (int i = 0; i < this.headersHolder.Length; ++i) {
            final String columnValue = this.get(i);
            this.headersHolder.Headers[i] = columnValue;
            this.headersHolder.IndexByName.put(columnValue, new Integer(i));
        }
        if (result) {
            --this.currentRecord;
        }
        this.columnsCount = 0;
        return result;
    }
    
    public String getHeader(final int columnIndex) throws IOException {
        this.checkClosed();
        if (columnIndex > -1 && columnIndex < this.headersHolder.Length) {
            return this.headersHolder.Headers[columnIndex];
        }
        return "";
    }
    
    public boolean isQualified(final int columnIndex) throws IOException {
        this.checkClosed();
        return columnIndex < this.columnsCount && columnIndex > -1 && this.isQualified[columnIndex];
    }
    
    private void endColumn() throws IOException {
        String currentValue = "";
        if (this.startedColumn) {
            if (this.columnBuffer.Position == 0) {
                if (this.dataBuffer.ColumnStart < this.dataBuffer.Position) {
                    int lastLetter = this.dataBuffer.Position - 1;
                    if (this.userSettings.TrimWhitespace && !this.startedWithQualifier) {
                        while (lastLetter >= this.dataBuffer.ColumnStart && (this.dataBuffer.Buffer[lastLetter] == ' ' || this.dataBuffer.Buffer[lastLetter] == '\t')) {
                            --lastLetter;
                        }
                    }
                    currentValue = new String(this.dataBuffer.Buffer, this.dataBuffer.ColumnStart, lastLetter - this.dataBuffer.ColumnStart + 1);
                }
            }
            else {
                this.updateCurrentValue();
                int lastLetter = this.columnBuffer.Position - 1;
                if (this.userSettings.TrimWhitespace && !this.startedWithQualifier) {
                    while (lastLetter >= 0 && (this.columnBuffer.Buffer[lastLetter] == ' ' || this.columnBuffer.Buffer[lastLetter] == ' ')) {
                        --lastLetter;
                    }
                }
                currentValue = new String(this.columnBuffer.Buffer, 0, lastLetter + 1);
            }
        }
        this.columnBuffer.Position = 0;
        this.startedColumn = false;
        if (this.columnsCount >= 100000 && this.userSettings.SafetySwitch) {
            this.close();
            throw new IOException("Maximum column count of 100,000 exceeded in record " + NumberFormat.getIntegerInstance().format(this.currentRecord) + ". Set the SafetySwitch property to false" + " if you're expecting more than 100,000 columns per record to" + " avoid this error.");
        }
        if (this.columnsCount == this.values.length) {
            final int newLength = this.values.length * 2;
            final String[] holder = new String[newLength];
            System.arraycopy(this.values, 0, holder, 0, this.values.length);
            this.values = holder;
            final boolean[] qualifiedHolder = new boolean[newLength];
            System.arraycopy(this.isQualified, 0, qualifiedHolder, 0, this.isQualified.length);
            this.isQualified = qualifiedHolder;
        }
        this.values[this.columnsCount] = currentValue;
        this.isQualified[this.columnsCount] = this.startedWithQualifier;
        currentValue = "";
        ++this.columnsCount;
    }
    
    private void appendLetter(final char letter) {
        if (this.columnBuffer.Position == this.columnBuffer.Buffer.length) {
            final int newLength = this.columnBuffer.Buffer.length * 2;
            final char[] holder = new char[newLength];
            System.arraycopy(this.columnBuffer.Buffer, 0, holder, 0, this.columnBuffer.Position);
            this.columnBuffer.Buffer = holder;
        }
        this.columnBuffer.Buffer[this.columnBuffer.Position++] = letter;
        this.dataBuffer.ColumnStart = this.dataBuffer.Position + 1;
    }
    
    private void updateCurrentValue() {
        if (this.startedColumn && this.dataBuffer.ColumnStart < this.dataBuffer.Position) {
            if (this.columnBuffer.Buffer.length - this.columnBuffer.Position < this.dataBuffer.Position - this.dataBuffer.ColumnStart) {
                final int newLength = this.columnBuffer.Buffer.length + Math.max(this.dataBuffer.Position - this.dataBuffer.ColumnStart, this.columnBuffer.Buffer.length);
                final char[] holder = new char[newLength];
                System.arraycopy(this.columnBuffer.Buffer, 0, holder, 0, this.columnBuffer.Position);
                this.columnBuffer.Buffer = holder;
            }
            System.arraycopy(this.dataBuffer.Buffer, this.dataBuffer.ColumnStart, this.columnBuffer.Buffer, this.columnBuffer.Position, this.dataBuffer.Position - this.dataBuffer.ColumnStart);
            final ColumnBuffer columnBuffer = this.columnBuffer;
            columnBuffer.Position += this.dataBuffer.Position - this.dataBuffer.ColumnStart;
        }
        this.dataBuffer.ColumnStart = this.dataBuffer.Position + 1;
    }
    
    private void endRecord() throws IOException {
        this.hasReadNextLine = true;
        ++this.currentRecord;
    }
    
    public int getIndex(final String headerName) throws IOException {
        this.checkClosed();
        final Object indexValue = this.headersHolder.IndexByName.get(headerName);
        if (indexValue != null) {
            return (int)indexValue;
        }
        return -1;
    }
    
    public boolean skipRecord() throws IOException {
        this.checkClosed();
        boolean recordRead = false;
        if (this.hasMoreData) {
            recordRead = this.readRecord();
            if (recordRead) {
                --this.currentRecord;
            }
        }
        return recordRead;
    }
    
    public boolean skipLine() throws IOException {
        this.checkClosed();
        this.columnsCount = 0;
        boolean skippedLine = false;
        if (this.hasMoreData) {
            boolean foundEol = false;
            do {
                if (this.dataBuffer.Position == this.dataBuffer.Count) {
                    this.checkDataLength();
                }
                else {
                    skippedLine = true;
                    final char currentLetter = this.dataBuffer.Buffer[this.dataBuffer.Position];
                    if (currentLetter == '\r' || currentLetter == '\n') {
                        foundEol = true;
                    }
                    this.lastLetter = currentLetter;
                    if (foundEol) {
                        continue;
                    }
                    final DataBuffer dataBuffer = this.dataBuffer;
                    ++dataBuffer.Position;
                }
            } while (this.hasMoreData && !foundEol);
            this.columnBuffer.Position = 0;
            this.dataBuffer.LineStart = this.dataBuffer.Position + 1;
        }
        this.rawBuffer.Position = 0;
        this.rawRecord = "";
        return skippedLine;
    }
    
    public void close() {
        if (!this.closed) {
            this.close(true);
            this.closed = true;
        }
    }
    
    private void close(final boolean closing) {
        if (!this.closed) {
            if (closing) {
                this.charset = null;
                this.headersHolder.Headers = null;
                this.headersHolder.IndexByName = null;
                this.dataBuffer.Buffer = null;
                this.columnBuffer.Buffer = null;
                this.rawBuffer.Buffer = null;
            }
            try {
                if (this.initialized) {
                    this.inputStream.close();
                }
            }
            catch (final Exception ex) {}
            this.inputStream = null;
            this.closed = true;
        }
    }
    
    private void checkClosed() throws IOException {
        if (this.closed) {
            throw new IOException("This instance of the CsvReader class has already been closed.");
        }
    }
    
    @Override
    protected void finalize() {
        this.close(false);
    }
    
    private static char hexToDec(final char hex) {
        char result;
        if (hex >= 'a') {
            result = (char)(hex - 'a' + 10);
        }
        else if (hex >= 'A') {
            result = (char)(hex - 'A' + 10);
        }
        else {
            result = (char)(hex - '0');
        }
        return result;
    }
    
    private class ComplexEscape
    {
        private static final int UNICODE = 1;
        private static final int OCTAL = 2;
        private static final int DECIMAL = 3;
        private static final int HEX = 4;
    }
    
    private class DataBuffer
    {
        public char[] Buffer;
        public int Position;
        public int Count;
        public int ColumnStart;
        public int LineStart;
        
        public DataBuffer() {
            this.Buffer = new char[1024];
            this.Position = 0;
            this.Count = 0;
            this.ColumnStart = 0;
            this.LineStart = 0;
        }
    }
    
    private class ColumnBuffer
    {
        public char[] Buffer;
        public int Position;
        
        public ColumnBuffer() {
            this.Buffer = new char[50];
            this.Position = 0;
        }
    }
    
    private class RawRecordBuffer
    {
        public char[] Buffer;
        public int Position;
        
        public RawRecordBuffer() {
            this.Buffer = new char[500];
            this.Position = 0;
        }
    }
    
    private class Letters
    {
        public static final char LF = '\n';
        public static final char CR = '\r';
        public static final char QUOTE = '\"';
        public static final char COMMA = ',';
        public static final char SPACE = ' ';
        public static final char TAB = '\t';
        public static final char POUND = '#';
        public static final char BACKSLASH = '\\';
        public static final char NULL = '\0';
        public static final char BACKSPACE = '\b';
        public static final char FORM_FEED = '\f';
        public static final char ESCAPE = '\u001b';
        public static final char VERTICAL_TAB = '\u000b';
        public static final char ALERT = '\u0007';
    }
    
    private class UserSettings
    {
        public boolean CaseSensitive;
        public char TextQualifier;
        public boolean TrimWhitespace;
        public boolean UseTextQualifier;
        public char Delimiter;
        public char RecordDelimiter;
        public char Comment;
        public boolean UseComments;
        public int EscapeMode;
        public boolean SafetySwitch;
        public boolean SkipEmptyRecords;
        public boolean CaptureRawRecord;
        
        public UserSettings() {
            this.CaseSensitive = true;
            this.TextQualifier = '\"';
            this.TrimWhitespace = true;
            this.UseTextQualifier = true;
            this.Delimiter = ',';
            this.RecordDelimiter = '\0';
            this.Comment = '#';
            this.UseComments = false;
            this.EscapeMode = 1;
            this.SafetySwitch = true;
            this.SkipEmptyRecords = true;
            this.CaptureRawRecord = true;
        }
    }
    
    private class HeadersHolder
    {
        public String[] Headers;
        public int Length;
        public HashMap IndexByName;
        
        public HeadersHolder() {
            this.Headers = null;
            this.Length = 0;
            this.IndexByName = new HashMap();
        }
    }
    
    private class StaticSettings
    {
        public static final int MAX_BUFFER_SIZE = 1024;
        public static final int MAX_FILE_BUFFER_SIZE = 4096;
        public static final int INITIAL_COLUMN_COUNT = 10;
        public static final int INITIAL_COLUMN_BUFFER_SIZE = 50;
    }
}
